"""
This module contains the wrapper for `Quart.request` that can be used inside web methods
"""

from quart import request
import functools


class EasyRequest:
    """ A helper class that makes request handling more uniform.

    All attributes are assessable via the same syntax, while with Flask.request
    or quart.request, you will have slightly different syntax when retrieving
    different request attributes, which can be very irritating and take much
    time debugging your code for missing parentheses or an await statement.

    .. hint::
        You only need to await attributes that need calculation, for example
        evaluating the request body, like `.json` or `.form`.

        .. code-block:: python

            headers = EasyRequest.headers
            params = EasyRequest.params
            form = await EasyRequest.form # requires time for calculation
            json = await EasyRequest.json # requires time for calculation
    """

    params: dict = ...
    """ The URL parameters like Flask.request.params. """

    headers: dict = ...
    """ The request headers like quart.request.headers. """

    cookies: dict = ...
    """ The request cookies like quart.request.cookies. """

    __quart_request: request = ...
    """ The quart.request. instance that is used in the current scope. """

    def __load(self):
        """ Loads the content of quart.request into this instance and returns it. """

        self.__quart_request = request
        self.params = dict(self.__quart_request.args)
        self.headers = dict(self.__quart_request.headers)
        self.cookies = dict(self.__quart_request.cookies)
        return self

    def __call__(self, f):
        """ Decorates an endpoint function to use the `EasyRequest` with, which makes
        `EasyRequest` available in that endpoint method.
        """

        @functools.wraps(f)
        async def decorated_function(*args2, **kwargs2):
            f.__globals__[self.__class__.__name__] = self.__class__().__load()
            return await f(*args2, **kwargs2)

        return decorated_function

    @property
    async def form(self):
        """ The request form data like quart.request.form. """
        return dict(await self.__quart_request.form)

    @property
    async def json(self):
        """ The request body data (as JSON) like quart.request.form.

        Be aware that in order for quart.request.get_json() to return
        a JSON dictionary, the `Content-Type` header must be set to
        `application/json`.
        """
        json = await self.__quart_request.get_json()
        return {} if json is None else json


if __name__ == '__main__':
    from Aeros import WebServer
    from quart import jsonify, render_template
    from time import strftime

    app = WebServer(__name__, host="0.0.0.0", port=80)  # init backend & web server
    app.template_folder = "./frontend/build/"
    app.static_folder = "./frontend/build/static/"


    @app.route("/")
    async def home():
        return await render_template("index.html")


    @app.route("/time/")
    async def current_time():
        print("Headers:", EasyRequest.headers)
        print("Params :", EasyRequest.params)
        print("Cookies:", EasyRequest.cookies)
        print("Form   :", await EasyRequest.form)
        print("JSON   :", await EasyRequest.json)
        return jsonify({"timestamp": strftime("%H:%M:%S")})


    app.run()  # run web server in this thread (endless mode)



"""
This module contains features for server-side caching by wrapping the `flask_caching` module.
"""

from flask_caching import Cache
import warnings


class SimpleCache(Cache):
    """ The most basic cache that requires no set-up and no arguments. """

    def __new__(cls, *args, **kwargs):
        return Cache(
            *args, **kwargs,
            config={
                'CACHE_TYPE': cls.__name__
            })


class FileSystemCache(Cache):
    """ Stores cached responses on the file system,
    good for huge caches that don't fit into memory."""

    def __new__(cls, directory: str, *args, **kwargs):
        return Cache(
            *args, **kwargs,
            config={
                'CACHE_TYPE': cls.__name__,
                'CACHE_DIR': directory
            })


# For backwards compatibility
class FilesystemCache:
    """ An alias for `FileSystemCache`

    .. warning::
        This class is deprecated! Use `FileSystemCache` instead.
    """

    def __new__(cls, *args, **kwargs):
        warnings.warn('"FilesystemCache" is deprecated! Use "FileSystemCache" instead.')
        return FileSystemCache(
            *args, **kwargs
        )


class RedisCache(Cache):
    """ A Redis client to store responses on an external Redis server. """

    # ignore different signature
    def __new__(cls, host: str, port: int, user: str, password: str, db: str, *args, **kwargs):
        return Cache(
            *args, **kwargs,
            config={
                'CACHE_TYPE': cls.__name__,
                'CACHE_REDIS_HOST': host,
                'CACHE_REDIS_PORT': port,
                'CACHE_REDIS_PASSWORD': password,
                'CACHE_REDIS_DB': db
            })



"""
The compression module includes classes for compression algorithms implemented by `quart_compress`.
These are compatible with the async coroutine framework that Aeros uses and will compress the web
server's responses before sending it to the client.
"""

from typing import List
from quart_compress import Compress
from quart import Quart
import warnings


class Base:
    """
    The base class that implements the wrapper for `quart_compress.Compress`.
    """
    algorithm = 'gzip'

    def __init__(self, level: int = 2, min_size: int = 500, mimetypes: List = None):
        self.compressor = Compress()
        self.level = level
        self.min_size = min_size
        self.mimetypes = mimetypes if mimetypes else ['text/plain', 'text/html', 'text/css', 'text/scss', 'text/xml', 'application/json', 'application/javascript']

    def init_app(self, app: Quart):
        """
        Initializes the `Compress` instance with the give `WebServer`/`Quart` application instance

        Arguments:
            app (Quart): The web server instance to use the caching functionality with
        """
        app.config["COMPRESS_ALGORITHM"] = self.algorithm
        app.config["COMPRESS_MIN_SIZE"] = self.min_size
        app.config["COMPRESS_LEVEL"] = self.level
        app.config["COMPRESS_MIMETYPES"] = self.mimetypes
        print("Algorithm is", self.algorithm)
        self.compressor.init_app(app)


class Gzip(Base):
    """
    This class should be used to implement a Gzip compression to your Aeros web server.
    """
    algorithm = 'gzip'


class Br(Base):
    """
    This class should be used to implement a Br compression to your Aeros web server.
    """
    algorithm = 'br'


class Compression:
    """
    The former class that represented compression for Aeros. Falls back to `Gzip`.

    .. warning::
        This class is deprecated and should not be used. Use `Gzip` or `Br` instead.
    """

    def __new__(cls, *args, **kwargs):
        warnings.warn('"Compression" is deprecated. Please use "Gzip" or "Br" instead.')
        return Gzip(*args, **kwargs)



"""
This module contains features for server-side and client-side caching.
"""

from .server import (
    SimpleCache,
    FileSystemCache,
    FilesystemCache,
    RedisCache
)
from .client import (
    CacheControl,
    CacheTypes
)



"""
This module contains features for client-side caching using the Cache-Control header.
"""

from enum import Enum
from quart import Response
import functools
from quart import make_response


class CacheTypes(Enum):
    """
    An enum to replace the cache option strings with variables for auto-complete in most IDEs.
    """
    NO_CACHE = 'no-cache'
    NO_STORE = 'no-store'
    PUBLIC = 'public'
    PRIVATE = 'private'


class CacheControl:
    """
    This class manipulates the Cache-Control header in all responses sent from an endpoint
    decorated with this class.

    .. hint::
        If you need conditional caching rules you may still use `response` from quart to
        manipulate the Cache-Control header to your needs.
    """

    def __init__(
            self, cache_type: CacheTypes, max_age: int = None,
            immutable: bool = False, no_transform: bool = False,
            stale_while_revalidate: int = None, stale_if_error: bool = False,
            cache_on_status: list = [200]
    ):
        """
        Arguments:
            cache_type (CacheTypes): public | private | no-cache | no-store
            max_age (int): mag-age in seconds
            no_transform (bool): disallow proxy changes?
            immutable (bool): If set, browsers will not re-validate
            stale_while_revalidate (int): Time in seconds how long we allow to serve the old content while re-validating
            stale_if_error (bool): Serve the last cached content if backend gives error code
            cache_on_status (list): A list of HTTP response codes when to apply the policy (default: [200])

        """
        self.cache_on_status = cache_on_status
        self.params = [cache_type.value]
        self.params.append(f'max_age={max_age}') if max_age else None,
        self.params.append('immutable') if immutable else None
        self.params.append('no-transform') if no_transform else None
        self.params.append(f'stale-while-revalidate={stale_while_revalidate}') if stale_while_revalidate else None
        self.params.append('stale-if-error') if stale_if_error else None

    def __call__(self, f: callable) -> callable:
        @functools.wraps(f)
        async def wrapped(*args, **kwargs):
            response = await f(*args, **kwargs)

            if not isinstance(response, Response):
                response = await make_response(response)

            if response.status_code in self.cache_on_status:
                response.headers['cache-control'] = ', '.join(self.params)

            return response

        return wrapped



"""
The misc module contains useful helper functions and decorators to simplify development and make
the resulting code much more readable and clean.
"""

from asgiref.sync import sync_to_async, async_to_sync



"""
This module contains threading helper classes and functions used to coordinate
applications with multi-thread support.
"""

import threading
import ctypes
import warnings


class AdvancedThread(threading.Thread):
    """ This thread class extends the pure-python "Thread" class
    by a stop() function to terminate the thread's run() method.

    .. warning::
        The stop feature is still experimental and needs extensive testing.
    """

    def __init__(self, *args, **kwargs):
        threading.Thread.__init__(self, *args, **kwargs)

    def __get_id(self):
        """ Get's the threads PID. """

        if hasattr(self, '_thread_id'):
            return self._thread_id
        for id, thread in threading._active.items():
            if thread is self:
                return id

    def stop(self):
        """ Sends a kill signal to the given thread """

        warnings.warn('"AdvancedThread.stop()" is still experimental. Use with caution.')
        thread_id = self.__get_id()
        res = ctypes.pythonapi.PyThreadState_SetAsyncExc(thread_id, ctypes.py_object(SystemExit))
        if res > 1:
            ctypes.pythonapi.PyThreadState_SetAsyncExc(thread_id, 0)
            print('Exception raise failure')



import setuptools

with open("requirements.txt", "r") as fh:
    requirements = [x.strip("\n") for x in fh.readlines()]

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="Aeros",
    version="2.0.0b2",
    author="TheClockTwister",
    description="High-performance ASGI framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/TheClockTwister/Aeros",
    packages=setuptools.find_packages(),
    install_requires=requirements,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Framework :: AsyncIO",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',  # required for f"" string format notation
)



"""
This module contains the main web server class
"""

import functools
import inspect
from quart import Quart
import hashlib
import uvicorn
import logging
import logging.handlers
from logging import INFO
import sys

from .caching.server import Cache
from .compression import Base
from .Request import EasyRequest


class WebServer(Quart):
    """ This is the main server class which extends a standard Flask class by a bunch of features and major
    performance improvements. It extends the quart class, which by itself is already an enhanced version of
    the Flask class. This class however allows production-grade deployment using the uvicorn ASGI server
    as production server. But instead of calling the uvicorn command via the console, it can be started
    directly from the Python code itself, making it easier to integrate in higher-level scripts and
    applications without calling os.system() or subprocess.Popen().
    """

    __log_format_std = '[%(asctime)s] %(levelname)s: %(message)s'  # for instance logger "quart.app"

    def __init__(self, import_name: str, host: str = "0.0.0.0", port: int = 80, include_server_header: bool = True,
                 logging_level: int = INFO, cache: Cache = None, compression: Base = None,
                 global_headers: dict = {}, *args, **kwargs):

        super().__init__(import_name, *args, **kwargs)

        self.log_level = logging_level
        self.logger.setLevel(self.log_level)
        self._host, self._port = host, port

        self._cache = cache
        self._compression = compression

        h = logging.StreamHandler(sys.stdout)  # make own logging format the same as uvicorn uses
        h.setFormatter(logging.Formatter(self.__log_format_std))
        self.logger.handlers = [h]

        self._global_headers = global_headers
        if include_server_header:
            self._global_headers['server'] = self._global_headers.get('server', 'Aeros 2.0.0 (uvicorn)')  # change server header to Aeros
        self._global_headers = list(global_headers.items())  # convert dict to list of tuples for uvicorn

    def cache(self, timeout=None, key_prefix="view/%s", unless=None, forced_update=None,
              response_filter=None, query_string=False, hash_method=hashlib.md5, cache_none=False, ):
        """ A simple wrapper that forwards the cached() decorator to the internal Cache() instance.
        May be used as the normal @cache.cached() decorator."""

        def decorator(f):
            if self._cache is None:
                print('No cache specified in', self)
                return f

            @functools.wraps(f)
            @self._cache.cached(timeout=timeout, key_prefix=key_prefix, unless=unless, forced_update=forced_update,
                                response_filter=response_filter, query_string=query_string, hash_method=hash_method, cache_none=cache_none)
            async def decorated_function(*args2, **kwargs2):
                x = await f(*args2, **kwargs2)
                return x

            return decorated_function

        return decorator

    def clear_cache(self):
        """ Clears the WebServer instance cache. Use with caution!"""
        self._cache.clear()

    def _get_own_instance_path(self) -> str:
        """
        Retrieves the file and variable name of this instance to be used in the uvicorn CLI.

        Since uvicorn needs the application's file and global variable name for multicore execution, an
        instance needs to know it's own origin file and name. But since this class is not defined in the
        same file as it is called or defined from, this method searches for the correct module/file and
        evaluates it's instance name.

        .. warning::
            This method is a beta feature. Use with caution (good testing).

         """
        for i in range(10):
            try:
                frame = inspect.stack()[i][0].f_globals.items()
                for key, val in frame:
                    try:
                        if val == self:
                            path = f"{dict(frame)['__name__']}:{key}"
                            self.logger.debug(f"{self.__class__.__name__} resolved itself to \"{path}\"")
                            return path
                    except (RuntimeError, AttributeError, KeyError):
                        pass
            except IndexError:
                self.logger.critical(f"{self.__class__.__name__} is unable to find own instance file:name. Launching in single-thread mode!")
                return None

    def run_server(self,
                   host: str = None, port: int = None, log_level: int = None,  # overrides for __init__
                   access_log_file: str = None, access_log_to_std: bool = True,
                   soft_limit: int = 32, hard_limit: int = 42,
                   **kwargs
                   ) -> None:
        """ Generates the necessary config and runs the server instance. Possible keyword arguments are the same as supported by the
            `uvicorn.run()` method as they are passed into `Config(app, **kwargs)`. This method also configures features like caching
            and compression that are not default in Quart or Flask and unique to Aeros or require third-party modules to be configured.
        """
        if kwargs.get('suppress_deprecation_warning', False):
            del kwargs['suppress_deprecation_warning']
        else:
            self.logger.warning('The usage of run_server() is deprecated. Use run() instead.')  # delete, otherwise it gets into the config

        # Initialize extra features just in case the user replaced them with their own instances
        if issubclass(self._cache.__class__, Cache):
            self._cache.init_app(self)
            self.logger.info("Caching is enabled")
        else:
            self.logger.info("Caching is disabled")

        if issubclass(self._compression.__class__, Base):
            self._compression.init_app(self)
            self.logger.info("Compression is enabled")
        else:
            self.logger.info("Compression is disabled")

        # create uvicorn configuration ------------------------------------------------------------------------------------------------

        config = {
            'app': self,
            'host': host if host else self._host,
            'port': port if port else self._port,
            'headers': kwargs.get('headers', self._global_headers),

            'log_level': log_level if log_level else self.log_level,
            'access_log': bool(access_log_file) or access_log_to_std,

            'limit_concurrency': soft_limit + 1,
            'backlog': hard_limit,

            **kwargs
        }

        # set logging information
        self.logger.setLevel(log_level if log_level else self.log_level)  # re-set log level for instance logger if changed
        config['log_config'] = {  # logging configuration for uvicorn.run()
            'version': 1, 'disable_existing_loggers': True,
            'formatters': {
                'default': {'()': 'logging.Formatter', 'fmt': self.__log_format_std},
                'access_file': {'()': 'uvicorn.logging.AccessFormatter', 'fmt': '[%(asctime)s] %(client_addr)s | "%(request_line)s" | %(status_code)s'},
                'access_std': {'()': 'uvicorn.logging.AccessFormatter', 'fmt': '[%(asctime)s] ACCESS: %(client_addr)s | "%(request_line)s" | %(status_code)s'}
            },
            'handlers': {
                'default': {'formatter': 'default', 'class': 'logging.StreamHandler', 'stream': 'ext://sys.stdout'},
                'error': {'formatter': 'default', 'class': 'logging.StreamHandler', 'stream': 'ext://sys.stderr'}
            },
            'loggers': {
                'uvicorn': {'level': log_level, 'handlers': ['default']},
                'uvicorn.access': {'level': 'INFO', 'propagate': False, 'handlers': []}
            }
        }

        if access_log_to_std:
            config['log_config']['handlers']['access_std'] = {'formatter': 'access_std', 'class': 'logging.StreamHandler', 'stream': 'ext://sys.stdout'}
            config['log_config']['loggers']['uvicorn.access']['handlers'].append('access_std')
        if access_log_file:
            config['log_config']['handlers']['access_file'] = {'formatter': 'access_file', 'class': 'logging.FileHandler', 'filename': access_log_file}
            config['log_config']['loggers']['uvicorn.access']['handlers'].append('access_file')

            # try multi-core execution
        if config.get('workers', None):
            self.logger.debug(f"{self.__class__.__name__}.run_server() multi-core execution is a beta feature. You can also use uvicorn CLI directly.")
            instance_path = self._get_own_instance_path()
            if instance_path is not None:
                # if instance found, run multi-core by replacing instance with path
                config['app'] = instance_path
                self.logger.info(f"Starting in multi-thread mode...")
            else:
                # if instance not found, delete worker count and launch single-thread
                del config['workers']
                self.logger.info(f"Starting in single-thread mode...")

        uvicorn.run(**config)

    def run(self, *args, **kwargs):
        return self.run_server(*args, **kwargs, suppress_deprecation_warning=True)

    def route(self, *args, **kwargs):
        def new_route_decorator(func):
            new_func = EasyRequest()(func)  # inject EasyRequest into every function so that it is always available
            Quart.route(super(WebServer, self), *args, **kwargs)(new_func)  # route the new function to Quart.route

        return new_route_decorator
